# Import the necessary modules
import cv2
import numpy as np
import sys

# Define a function to rescale an image to a size comparable to VGA size (480x600)
def rescale_image(image):
    # Get the original height and width of the image
    height, width = image.shape[:2]
    # Calculate the scaling factor based on the aspect ratio and the desired height
    scale = 480 / height
    # Resize the image using the scaling factor
    resized_image = cv2.resize(image, (int(width * scale), int(height * scale)))
    # Return the resized image
    return resized_image

# Define a function to extract SIFT keypoints and descriptors from an image
def extract_sift(image):
    # Convert the image to grayscale
    gray = cv2.cvtColor(image, cv2.COLOR_BGR2GRAY)
    # Create a SIFT object
    sift = cv2.SIFT_create()
    # Detect and compute SIFT keypoints and descriptors
    keypoints, descriptors = sift.detectAndCompute(gray, None)
    # Return the keypoints and descriptors
    return keypoints, descriptors

# Define a function to draw keypoints on an image
def draw_keypoints(image, keypoints):
    # Create a copy of the image
    image_copy = image.copy()
    # Draw a cross and a circle for each keypoint
    for keypoint in keypoints:
        # Get the keypoint coordinates, size and angle
        x, y = keypoint.pt
        s = keypoint.size
        a = keypoint.angle
        # Convert the angle to radians
        a = np.deg2rad(a)
        # Calculate the end points of the cross
        x1 = int(x - s * np.sin(a))
        y1 = int(y + s * np.cos(a))
        x2 = int(x + s * np.sin(a))
        y2 = int(y - s * np.cos(a))
        # Draw a line for the cross
        cv2.line(image_copy, (x1, y1), (x2, y2), (0, 0, 255), 2)
        # Calculate the end point of the orientation line
        x3 = int(x + s * np.cos(a))
        y3 = int(y + s * np.sin(a))
        # Draw a line for the orientation
        cv2.line(image_copy, (int(x), int(y)), (x3, y3), (0, 255, 0), 2)
        # Draw a circle for the scale
        cv2.circle(image_copy, (int(x), int(y)), int(s), (255, 0, 0), 2)
    # Return the image with keypoints
    return image_copy

# Define a function to cluster SIFT descriptors into K clusters using K-means algorithm
def cluster_sift(descriptors_list, K):
    # Concatenate all descriptors into a single array
    descriptors_array = np.concatenate(descriptors_list, axis=0)
    # Convert the array to float32 type
    descriptors_array = descriptors_array.astype(np.float32)
    # Define the criteria for K-means clustering (termination condition and number of iterations)
    criteria = (cv2.TERM_CRITERIA_EPS + cv2.TERM_CRITERIA_MAX_ITER, 10, 1.0)
    # Perform K-means clustering on the descriptors array
    _, labels, centers = cv2.kmeans(descriptors_array, K, None, criteria, 10, cv2.KMEANS_RANDOM_CENTERS)
    # Return the labels and centers of the clusters
    return labels, centers

# Define a function to construct a histogram of visual words for an image
def construct_histogram(descriptors, labels, K):
    # Initialize an empty histogram with K bins
    histogram = np.zeros(K)
    # Loop through each descriptor of the image
    for i in range(len(descriptors)):
        # Get the label of the cluster that the descriptor belongs to
        label = labels[i]
        # Increment the corresponding bin of the histogram by one
        histogram[label] += 1
    # Normalize the histogram by dividing each bin by the total number of descriptors
    histogram = histogram / len(descriptors)
    # Return the normalized histogram
    return histogram

# Define a function to calculate the distance between two histograms using chi-square distance metric
def calculate_distance(histogram1, histogram2):
    # Calculate the chi-square distance between two histograms using numpy formula 
    distance = np.sum((histogram1 - histogram2) ** 2 / (histogram1 + histogram2 + 1e-10))
    # Return the distance value
    return distance

# Get the input arguments from the command line
args = sys.argv[1:]

# Check if there is at least one input argument
if len(args) > 0:
    # Initialize an empty list to store the image names
    image_names = []
    # Initialize an empty list to store the rescaled images
    images = []
    # Initialize an empty list to store the keypoints of each image
    keypoints_list = []
    # Initialize an empty list to store the descriptors of each image
    descriptors_list = []
    # Loop through each input argument
    for arg in args:
        # Try to read the image file
        image = cv2.imread(arg)
        # Check if the image is valid
        if image is not None:
            # Append the image name to the image names list
            image_names.append(arg)
            # Rescale the image and append it to the images list
            rescaled_image = rescale_image(image)
            images.append(rescaled_image)
            # Extract SIFT keypoints and descriptors and append them to the corresponding lists
            keypoints, descriptors = extract_sift(rescaled_image)
            keypoints_list.append(keypoints)
            descriptors_list.append(descriptors)
        else:
            # Print an error message if the image is invalid
            print(f"Invalid image file: {arg}")
    
    # Check if there is only one valid image file
    if len(images) == 1:
        # Perform task one: draw keypoints on the image and display both images
        # Get the original and rescaled images
        original_image = images[0]
        rescaled_image = images[0]
        # Get the keypoints of the rescaled image
        keypoints = keypoints_list[0]
        # Draw keypoints on the rescaled image
        keypoints_image = draw_keypoints(rescaled_image, keypoints)
        # Display both images in a window
        cv2.imshow(f"Original Image: {image_names[0]}", original_image)
        cv2.imshow(f"Keypoints Image: {image_names[0]}", keypoints_image)
        cv2.waitKey(0)
        cv2.destroyAllWindows()
    
    # Check if there are more than one valid image files
    elif len(images) > 1:
        # Perform task two: compare each pair of images using a Bag-of-Words model
        # Define the value of K as a percentage of the total number of keypoints
        K = int(0.1 * sum([len(keypoints) for keypoints in keypoints_list]))
        # Cluster the SIFT descriptors into K clusters using K-means algorithm
        labels, centers = cluster_sift(descriptors_list, K)
        # Initialize an empty list to store the histograms of visual words for each image
        histograms = []
        # Initialize a variable to keep track of the starting index of each image's labels in the labels array
        start_index = 0
        # Loop through each image's descriptors and labels
        for i in range(len(images)):
            # Get the descriptors and labels of the current image
            descriptors = descriptors_list[i]
            end_index = start_index + len(descriptors)
            image_labels = labels[start_index:end_index]
            # Construct a histogram of visual words for the current image and append it to the histograms list
            histogram = construct_histogram(descriptors, image_labels, K)
            histograms.append(histogram)
            # Update the start index for the next image's labels
            start_index = end_index
        
        # Loop through each pair of images and calculate their distance using chi-square distance metric
        for i in range(len(images) - 1):
            for j in range(i + 1, len(images)):
                # Get the histograms of visual words for the pair of images
                histogram1 = histograms[i]
                histogram2 = histograms[j]
                # Calculate the distance between the histograms
                distance = calculate_distance(histogram1, histogram2)
                # Print the distance value along with the pair of image names
                print(f"Distance between {image_names[i]} and {image_names[j]}: {distance:.4f}")
    
    else:
        # Print an error message if there are no valid image files
        print("No valid image files found.")
else:
    # Print an error message if there are no input arguments
    print("Please provide one or more image files as input arguments.")
